@@ -8,11 +8,11 @@ static const MVMREPROps Decoder_this_repr;
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
     MVMSTable *st  = MVM_gc_allocate_stable(tc, &Decoder_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMDecoder);
-    });
+    }
 
     return st->WHAT;
 }
@@ -233,11 +233,13 @@ void MVM_decoder_add_bytes(MVMThreadContext *tc, MVMDecoder *decoder, MVMObject
             default:
                 MVM_exception_throw_adhoc(tc, "Can only add bytes from an int array to a decoder");
         }
-        copy = MVM_malloc(output_size);
-        memcpy(copy, output, output_size);
-        enter_single_user(tc, decoder);
-        MVM_string_decodestream_add_bytes(tc, ds, copy, output_size);
-        exit_single_user(tc, decoder);
+        if (output_size) {
+            copy = MVM_malloc(output_size);
+            memcpy(copy, output, output_size);
+            enter_single_user(tc, decoder);
+            MVM_string_decodestream_add_bytes(tc, ds, copy, output_size);
+            exit_single_user(tc, decoder);
+        }
     }
     else {
         MVM_exception_throw_adhoc(tc, "Cannot add bytes to a decoder with a %s",
@@ -251,9 +253,9 @@ MVMString * MVM_decoder_take_chars(MVMThreadContext *tc, MVMDecoder *decoder, MV
                                    MVMint64 eof) {
     MVMString *result = NULL;
     enter_single_user(tc, decoder);
-    MVMROOT(tc, decoder, {
+    MVMROOT(tc, decoder) {
         result = MVM_string_decodestream_get_chars(tc, get_ds(tc, decoder), (MVMint32)chars, eof);
-    });
+    }
     exit_single_user(tc, decoder);
     return result;
 }
@@ -262,9 +264,9 @@ MVMString * MVM_decoder_take_chars(MVMThreadContext *tc, MVMDecoder *decoder, MV
 MVMString * MVM_decoder_take_all_chars(MVMThreadContext *tc, MVMDecoder *decoder) {
     MVMString *result = NULL;
     enter_single_user(tc, decoder);
-    MVMROOT(tc, decoder, {
+    MVMROOT(tc, decoder) {
         result = MVM_string_decodestream_get_all(tc, get_ds(tc, decoder));
-    });
+    }
     exit_single_user(tc, decoder);
     return result;
 }
@@ -273,9 +275,9 @@ MVMString * MVM_decoder_take_all_chars(MVMThreadContext *tc, MVMDecoder *decoder
 MVMString * MVM_decoder_take_available_chars(MVMThreadContext *tc, MVMDecoder *decoder) {
     MVMString *result = NULL;
     enter_single_user(tc, decoder);
-    MVMROOT(tc, decoder, {
+    MVMROOT(tc, decoder) {
         result = MVM_string_decodestream_get_available(tc, get_ds(tc, decoder));
-    });
+    }
     exit_single_user(tc, decoder);
     return result;
 }
@@ -287,11 +289,11 @@ MVMString * MVM_decoder_take_line(MVMThreadContext *tc, MVMDecoder *decoder,
     MVMDecodeStreamSeparators *sep_spec = get_sep_spec(tc, decoder);
     MVMString *result = NULL;
     enter_single_user(tc, decoder);
-    MVMROOT(tc, decoder, {
+    MVMROOT(tc, decoder) {
         result = incomplete_ok
             ? MVM_string_decodestream_get_until_sep_eof(tc, ds, sep_spec, (MVMint32)chomp)
             : MVM_string_decodestream_get_until_sep(tc, ds, sep_spec, (MVMint32)chomp);
-    });
+    }
     exit_single_user(tc, decoder);
     return result;
 }
@@ -329,10 +331,12 @@ MVMObject * MVM_decoder_take_bytes(MVMThreadContext *tc, MVMDecoder *decoder,
     if (MVM_string_decodestream_bytes_available(tc, ds) < bytes)
         return tc->instance->VMNull;
 
-    result = MVM_repr_alloc_init(tc, buf_type);
     enter_single_user(tc, decoder);
     read = MVM_string_decodestream_bytes_to_buf(tc, ds, &buf, bytes);
     exit_single_user(tc, decoder);
+
+    /* Allocate after we're done with decoder to avoid having to MVMROOT */
+    result = MVM_repr_alloc_init(tc, buf_type);
     ((MVMArray *)result)->body.slots.i8 = (MVMint8 *)buf;
     ((MVMArray *)result)->body.start    = 0;
     ((MVMArray *)result)->body.ssize    = read;
