@@ -145,16 +145,30 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
     /* Otherwise, need to do the work; cache it if in range. */
     const int is_negative = i < 0;
     const int msb = 64 - __builtin_clzll((is_negative ? -i : i) | 1);
-    char *buffer = MVM_malloc(mag[msb] + is_negative + 1);
+    const int max_size = mag[msb] + is_negative + 1;
+    char *buffer;
+    MVMString *result;
+    if (max_size <= 8) {
+        result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
+        result->body.storage_type = MVM_STRING_IN_SITU_8;
+        buffer = (char *)result->body.storage.in_situ_8;
+    }
+    else {
+        buffer = MVM_malloc(max_size);
+    }
     const int len = i64toa_jeaiii(i, buffer) - buffer;
     if (0 <= len) {
-        MVMString *result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
+        if (max_size > 8)
+            result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
+        else
+            result->body.num_graphs = len;
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
         return result;
     }
     else {
-        MVM_free(buffer);
+        if (max_size > 8)
+            MVM_free(buffer);
         MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRId64")", i);
     }
 }
@@ -169,17 +183,31 @@ MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
     }
     /* Otherwise, need to do the work; cache it if in range. */
     const int msb = 64 - __builtin_clzll(i | 1);
-    char *buffer = MVM_malloc(mag[msb] + 1);
-    const int len = u64toa_jeaiii(i, buffer) - buffer;
+    const int max_size = mag[msb] + 1;
+    char *buffer;
+    MVMString *result;
+    if (max_size <= 8) {
+        result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
+        result->body.storage_type = MVM_STRING_IN_SITU_8;
+        buffer = (char *)result->body.storage.in_situ_8;
+    }
+    else {
+        buffer = MVM_malloc(max_size);
+    }
+    const int len = i64toa_jeaiii(i, buffer) - buffer;
     if (0 <= len) {
-        MVMString *result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
+        if (max_size > 8)
+            result = MVM_string_ascii_from_buf_nocheck(tc, (MVMGrapheme8 *)buffer, len);
+        else
+            result->body.num_graphs = len;
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
         return result;
     }
     else {
-        MVM_free(buffer);
-        MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRIu64")", i);
+        if (max_size > 8)
+            MVM_free(buffer);
+        MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRId64")", i);
     }
 }
 
@@ -780,17 +808,17 @@ MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMi
 
     /* initialize the object */
     result = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVMROOT(tc, result, {
+    MVMROOT(tc, result) {
         MVMObject *box_type = MVM_hll_current(tc)->int_box_type;
-        MVMROOT(tc, box_type, {
+        MVMROOT(tc, box_type) {
             MVMObject *boxed = MVM_repr_box_int(tc, box_type, value);
             MVM_repr_push_o(tc, result, boxed);
             boxed = MVM_repr_box_int(tc, box_type, chars_really_converted);
             MVM_repr_push_o(tc, result, boxed);
             boxed = MVM_repr_box_int(tc, box_type, pos);
             MVM_repr_push_o(tc, result, boxed);
-        });
-    });
+        }
+    }
 
     return result;
 }
