@@ -1243,31 +1243,19 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
  * specified type. Non-existing object lexicals produce NULL, expected
  * (for better or worse) by various things. Otherwise, an error is thrown
  * if it does not exist. Incorrect type always throws. */
-int MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister *r) {
+MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type) {
     MVMSpeshFrameWalker fw;
     MVM_spesh_frame_walker_init_for_outers(tc, &fw, tc->cur_frame);
     MVMRegister *res = MVM_frame_lexical_lookup_using_frame_walker(tc, &fw, name, type);
 
-    if (res == NULL) {
-        MVMCode *resolver = tc->cur_frame->static_info->body.cu->body.resolver;
-        if (resolver) {
-            MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
-                    MVM_callsite_get_common(tc, MVM_CALLSITE_ID_STR));
-            args_record->args.source[0].s = name;
-            MVM_frame_dispatch_from_c(tc, resolver, args_record, r, MVM_RETURN_OBJ);
-        }
-        else if (MVM_UNLIKELY(type != MVM_reg_obj)) {
-            char *c_name = MVM_string_utf8_encode_C_string(tc, name);
-            char *waste[] = { c_name, NULL };
-            MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
-                c_name);
-        }
-        return 0;
-    }
-    else {
-        *r = *res;
-        return 1;
+    if (res == NULL && MVM_UNLIKELY(type != MVM_reg_obj)) {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
+            c_name);
     }
+
+    return res;
 }
 
 /* Binds the specified value to the given lexical, finding it along the static
@@ -1308,12 +1296,14 @@ MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *
 }
 
 /* Finds a lexical in the outer frame, throwing if it's not there. */
-void MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name, MVMRegister *result) {
-    int found;
+MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name) {
+    MVMRegister *r;
     MVMROOT(tc, name) {
-        found = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer, result);
-    }
-    if (MVM_UNLIKELY(!found)) {
+        r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
+    };
+    if (MVM_LIKELY(r != NULL))
+        return r->o;
+    else {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
@@ -1323,7 +1313,7 @@ void MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString *name,
 
 /* Looks up the address of the lexical with the specified name, starting with
  * the specified frame. Only works if it's an object lexical.  */
-int MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame, MVMRegister *r) {
+MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
     while (cur_frame != NULL) {
         if (cur_frame->static_info->body.num_lexicals) {
             MVMuint32 idx = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
@@ -1332,8 +1322,7 @@ int MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MV
                     MVMRegister *result = &cur_frame->env[idx];
                     if (!result->o)
                         MVM_frame_vivify_lexical(tc, cur_frame, idx);
-                    *r = *result;
-                    return 1;
+                    return result;
                 }
                 else {
                     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -1346,15 +1335,7 @@ int MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MV
         }
         cur_frame = cur_frame->outer;
     }
-    MVMCode *resolver = tc->cur_frame->static_info->body.cu->body.resolver;
-    if (resolver) {
-        MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
-                MVM_callsite_get_common(tc, MVM_CALLSITE_ID_STR));
-        args_record->args.source[0].s = name;
-        MVM_frame_dispatch_from_c(tc, resolver, args_record, r, MVM_RETURN_OBJ);
-        return 1;
-    }
-    return 0;
+    return NULL;
 }
 
 /* Performs some kind of lexical lookup using the frame walker. The exact walk
@@ -1529,8 +1510,8 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
             vivify, found_frame);
 }
 
-void MVM_frame_getdynlex_with_frame_walker(MVMThreadContext *tc, MVMSpeshFrameWalker *fw,
-                                                  MVMString *name, MVMRegister *r) {
+MVMObject * MVM_frame_getdynlex_with_frame_walker(MVMThreadContext *tc, MVMSpeshFrameWalker *fw,
+                                                  MVMString *name) {
     MVMuint16 type;
     MVMFrame *found_frame;
     MVMRegister *lex_reg = MVM_frame_find_dynamic_using_frame_walker(tc, fw, name, &type,
@@ -1593,26 +1574,12 @@ void MVM_frame_getdynlex_with_frame_walker(MVMThreadContext *tc, MVMSpeshFrameWa
                 MVM_exception_throw_adhoc(tc, "invalid register type in getdynlex: %d", type);
         }
     }
-    if (result) {
-        r->o = result;
-    }
-    else {
-        MVMCode *resolver = tc->cur_frame->static_info->body.cu->body.dynamic_resolver;
-        if (resolver) {
-            MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
-                    MVM_callsite_get_common(tc, MVM_CALLSITE_ID_STR));
-            args_record->args.source[0].s = name;
-            MVM_frame_dispatch_from_c(tc, resolver, args_record, r, MVM_RETURN_OBJ);
-        }
-        else {
-            r->o = tc->instance->VMNull;
-        }
-    }
+    return result ? result : tc->instance->VMNull;
 }
-void MVM_frame_getdynlex(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame, MVMRegister *r) {
+MVMObject * MVM_frame_getdynlex(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
     MVMSpeshFrameWalker fw;
     MVM_spesh_frame_walker_init(tc, &fw, cur_frame, 0);
-    MVM_frame_getdynlex_with_frame_walker(tc, &fw, name, r);
+    return MVM_frame_getdynlex_with_frame_walker(tc, &fw, name);
 }
 
 void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *value, MVMFrame *cur_frame) {
