@@ -24,8 +24,14 @@ MVM_STATIC_INLINE MVMuint32 MVM_str_hash_allocated_items(const struct MVMStrHash
     assert(!(control->cur_items == 0 && control->max_items == 0));
     return MVM_str_hash_official_size(control) + control->max_probe_distance_limit - 1;
 }
+/* Returns the number of buckets that the hash is using. This can be lower than
+ * the number of buckets allocated, as the last bucket that can be used is at
+ * max_probe_distance - 1, whereas buckets are allocated up to
+ * max_probe_distance_limit - 1. And if the hash is empty, it can't be using
+ * any buckets. As the name is meant to imply, this function is private. */
 MVM_STATIC_INLINE MVMuint32 MVM_str_hash_kompromat(const struct MVMStrHashTableControl *control) {
-    assert(!(control->cur_items == 0 && control->max_items == 0));
+    if (MVM_UNLIKELY(control->cur_items == 0))
+        return 0;
     return MVM_str_hash_official_size(control) + control->max_probe_distance - 1;
 }
 MVM_STATIC_INLINE MVMuint8 *MVM_str_hash_metadata(const struct MVMStrHashTableControl *control) {
@@ -94,7 +100,7 @@ MVM_STATIC_INLINE void MVM_str_hash_shallow_copy(MVMThreadContext *tc,
         MVM_oops(tc, "MVM_str_hash_shallow_copy called with a stale hashtable pointer");
     }
     if (control->cur_items == 0 && control->max_items == 0) {
-        struct MVMStrHashTableControl *empty = MVM_fixed_size_alloc(tc, tc->instance->fsa, sizeof(*empty));
+        struct MVMStrHashTableControl *empty = MVM_malloc(sizeof(*empty));
         memcpy(empty, control, sizeof(*empty));
         dest->table = empty;
     } else {
@@ -104,7 +110,7 @@ MVM_STATIC_INLINE void MVM_str_hash_shallow_copy(MVMThreadContext *tc,
         const char *start = (const char *)control - entries_size;
         size_t total_size
             = entries_size + sizeof(struct MVMStrHashTableControl) + metadata_size;
-        char *target = (char *) MVM_fixed_size_alloc(tc, tc->instance->fsa, total_size);
+        char *target = (char *) MVM_malloc(total_size);
         memcpy(target, start, total_size);
         dest->table = (struct MVMStrHashTableControl *)(target + entries_size);
     }
@@ -161,8 +167,7 @@ MVM_str_hash_create_loop_state(MVMThreadContext *tc,
     retval.probe_distance_shift = control->metadata_hash_bits;
     retval.max_probe_distance = control->max_probe_distance;
 
-    unsigned int used_hash_bits
-        = hash_val >> (control->key_right_shift - control->metadata_hash_bits);
+    unsigned int used_hash_bits = hash_val >> control->key_right_shift;
     retval.probe_distance = retval.metadata_increment | (used_hash_bits & retval.metadata_hash_mask);
     MVMHashNumItems bucket = used_hash_bits >> control->metadata_hash_bits;
     if (!control->metadata_hash_bits) {
