@@ -44,6 +44,9 @@ typedef struct {
     /* Did read already report EOF? */
     int eof_reported;
 
+    /* Did the open mode specify O_RDONLY, O_WRONLY, or O_RDWR? (ignores other mode settings) */
+    short open_mode;
+
     /* Output buffer, for buffered output. */
     char *output_buffer;
 
@@ -69,6 +72,8 @@ static MVMint64 mvm_fileno(MVMThreadContext *tc, MVMOSHandle *h) {
 /* Performs a write, either because a buffer filled or because we are not
  * buffering output. */
 static void perform_write(MVMThreadContext *tc, MVMIOFileData *data, char *buf, MVMint64 bytes) {
+    unsigned int interval_id = MVM_telemetry_interval_start(tc, "syncfile.perform_write");
+    MVMuint64 loops = 0;
     MVMint64 bytes_written = 0;
     MVM_gc_mark_thread_blocked(tc);
     while (bytes > 0) {
@@ -85,10 +90,15 @@ static void perform_write(MVMThreadContext *tc, MVMIOFileData *data, char *buf,
         bytes_written += r;
         buf += r;
         bytes -= r;
+        loops++;
     }
     MVM_gc_mark_thread_unblocked(tc);
     data->byte_position += bytes_written;
     data->known_writable = 1;
+    if (loops > 1)
+        MVM_telemetry_interval_annotate(loops, interval_id, "looped this many times");
+    MVM_telemetry_interval_annotate(bytes_written, interval_id, "this many bytes written");
+    MVM_telemetry_interval_stop((MVMThreadContext*)(uintptr_t)data->fd, interval_id, "syncfile.perform_write FD");
 }
 
 /* Flushes any existing output buffer and clears use back to 0. */
@@ -287,7 +297,7 @@ static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
     }
     MVM_gc_mark_thread_unblocked(tc);
 
-    MVM_exception_throw_adhoc(tc, "Failed to lock filehandle: %d", GetLastError());
+    MVM_exception_throw_adhoc(tc, "Failed to lock filehandle: %lu", GetLastError());
 
     return 0;
 
@@ -345,7 +355,7 @@ static void unlock(MVMThreadContext *tc, MVMOSHandle *h) {
     }
     MVM_gc_mark_thread_unblocked(tc);
 
-    MVM_exception_throw_adhoc(tc, "Failed to unlock filehandle: %d", GetLastError());
+    MVM_exception_throw_adhoc(tc, "Failed to unlock filehandle: %lu", GetLastError());
 #else
 
     struct flock l;
@@ -378,13 +388,26 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     }
 }
 
+/* Used to recover the MVM open mode from the (simple) POSIX-based open mode flag */
+static MVMint64 mvm_open_mode(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+
+    MVMint64 mvm_open_mode = -1;
+    switch(data->open_mode) {
+        case O_RDONLY: mvm_open_mode = 1; break;
+        case O_WRONLY: mvm_open_mode = 2; break;
+        case O_RDWR:   mvm_open_mode = 3; break;
+    }
+    return mvm_open_mode;
+}
+
 /* IO ops table, populated with functions. */
 static const MVMIOClosable      closable      = { closefh };
 static const MVMIOSyncReadable  sync_readable = { read_bytes, mvm_eof };
 static const MVMIOSyncWritable  sync_writable = { write_bytes, flush, truncatefh };
 static const MVMIOSeekable      seekable      = { seek, mvm_tell };
 static const MVMIOLockable      lockable      = { lock, unlock };
-static const MVMIOIntrospection introspection = { is_tty, mvm_fileno };
+static const MVMIOIntrospection introspection = { is_tty, mvm_fileno, mvm_open_mode };
 
 static const MVMIOOps op_table = {
     &closable,
@@ -403,8 +426,12 @@ static const MVMIOOps op_table = {
     gc_free
 };
 
-/* Builds POSIX flag from mode string. */
-static int resolve_open_mode(int *flag, const char *cp) {
+/*
+    Builds POSIX flag from mode string.
+    Because O_RDONLY is 0, we can't use bitwise logic on 'flag' when checking readonly/writeonly/readwrite.
+    So we store that initial determination in 'simple_flag'.
+*/
+static int resolve_open_mode(int *flag, short *simple_flag, const char *cp) {
     switch (*cp++) {
         case 'r': *flag = O_RDONLY; break;
         case '-': *flag = O_WRONLY; break;
@@ -418,6 +445,7 @@ static int resolve_open_mode(int *flag, const char *cp) {
         default:
         return 0;
     }
+    *simple_flag = *flag;
 
     for (;;) switch (*cp++) {
         case 0:
@@ -435,14 +463,15 @@ static int resolve_open_mode(int *flag, const char *cp) {
 
 /* Opens a file, returning a synchronous file handle. */
 MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMString *mode) {
-    char * const fname = MVM_string_utf8_c8_encode_C_string(tc, filename);
+    char * const fname = MVM_platform_path(tc, filename);
     int fd;
     int flag;
+    short simple_flag;
     STAT_t statbuf;
 
     /* Resolve mode description to flags. */
     char * const fmode  = MVM_string_utf8_encode_C_string(tc, mode);
-    if (!resolve_open_mode(&flag, fmode)) {
+    if (!resolve_open_mode(&flag, &simple_flag, fmode)) {
         char *waste[] = { fname, fmode, NULL };
         MVM_exception_throw_adhoc_free(tc, waste,
             "Invalid open mode for file %s: %s", fname, fmode);
@@ -484,6 +513,8 @@ MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMStrin
             tc->instance->boot_types.BOOTIO);
         data->fd          = fd;
         data->seekable    = MVM_platform_is_fd_seekable(fd);
+
+        data->open_mode   = simple_flag;
         result->body.ops  = &op_table;
         result->body.data = data;
         return (MVMObject *)result;
