@@ -236,8 +236,8 @@ static void discover_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *
     }
 }
 
-static void sp_guard_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
-                      MVMSpeshIns *ins) {
+void MVM_spesh_facts_guard_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
+        MVMSpeshIns *ins) {
     MVMuint16 opcode = ins->info->opcode;
 
     /* The sslot is always the second-to-last parameter, except for
@@ -246,15 +246,6 @@ static void sp_guard_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *b
 
     MVMSpeshFacts *facts    = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
 
-    /* We do not copy facts here, because it caused some trouble.
-     * Oftentimes, there were more flags set in the target register's
-     * facts before this, and got cleared by this copy operation.
-     * Also, it caused even the empty perl6 program to fail with a
-     * long stack trace about not finding an exception handler. */
-    /*copy_facts(tc, g,*/
-        /*ins->operands[0].reg.orig, ins->operands[0].reg.i,*/
-        /*ins->operands[1].reg.orig, ins->operands[1].reg.i);*/
-
     if (opcode == MVM_OP_sp_guard
             || opcode == MVM_OP_sp_guardconc
             || opcode == MVM_OP_sp_guardtype) {
@@ -452,7 +443,7 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             }
             ann = ann->next;
         }
-        if (p && ann_deopt_one && ann_logged && ins->info->opcode != MVM_OP_speshresolve)
+        if (p && ann_deopt_one && ann_logged)
             log_facts(tc, g, bb, ins, p, ann_deopt_one, ann_logged);
 
         /* Look for ops that are fact-interesting. */
@@ -469,8 +460,10 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             break;
         case MVM_OP_sp_fastcreate:
         case MVM_OP_sp_fastbox_i:
+        case MVM_OP_sp_fastbox_u:
         case MVM_OP_sp_fastbox_bi:
         case MVM_OP_sp_fastbox_i_ic:
+        case MVM_OP_sp_fastbox_u_ic:
         case MVM_OP_sp_fastbox_bi_ic:
         case MVM_OP_sp_add_I:
         case MVM_OP_sp_sub_I:
@@ -486,7 +479,8 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             break;
         case MVM_OP_box_s:
         case MVM_OP_box_i:
-        case MVM_OP_box_n: {
+        case MVM_OP_box_n:
+        case MVM_OP_box_u: {
                 MVMSpeshFacts *target_facts = &(g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i]);
                 create_facts(tc, g,
                     ins->operands[0].reg.orig, ins->operands[0].reg.i,
@@ -653,12 +647,17 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         case MVM_OP_getlexref_i32:
         case MVM_OP_getlexref_i16:
         case MVM_OP_getlexref_i8:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->int_lex_ref);
+            break;
+        case MVM_OP_getlexref_u:
         case MVM_OP_getlexref_u32:
         case MVM_OP_getlexref_u16:
         case MVM_OP_getlexref_u8:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
-                g->sf->body.cu->body.hll_config->int_lex_ref);
+                g->sf->body.cu->body.hll_config->uint_lex_ref);
             break;
         case MVM_OP_getlexref_n:
         case MVM_OP_getlexref_n32:
@@ -677,6 +676,12 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_attr_ref);
             break;
+        case MVM_OP_getattrref_u:
+        case MVM_OP_getattrsref_u:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->uint_attr_ref);
+            break;
         case MVM_OP_getattrref_n:
         case MVM_OP_getattrsref_n:
             create_facts_with_type(tc, g,
@@ -694,6 +699,11 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_pos_ref);
             break;
+        case MVM_OP_atposref_u:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->uint_pos_ref);
+            break;
         case MVM_OP_atposref_n:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
@@ -729,17 +739,13 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[4].reg.orig, ins->operands[4].reg.i);
             break;
-        case MVM_OP_setdispatcher:
-        case MVM_OP_setdispatcherfor:
-            g->sets_dispatcher = 1;
-            break;
         case MVM_OP_sp_guard:
         case MVM_OP_sp_guardconc:
         case MVM_OP_sp_guardtype:
         case MVM_OP_sp_guardobj:
         case MVM_OP_sp_guardjustconc:
         case MVM_OP_sp_guardjusttype:
-            sp_guard_facts(tc, g, bb, ins);
+            MVM_spesh_facts_guard_facts(tc, g, bb, ins);
             break;
         default:
             if (ins->info->opcode == (MVMuint16)-1)
@@ -765,6 +771,12 @@ static void tweak_block_handler_usage(MVMThreadContext *tc, MVMSpeshGraph *g) {
             operand.reg.i = 1;
             MVM_spesh_usages_add_for_handler_by_reg(tc, g, operand);
         }
+        if (g->sf->body.handlers[i].category_mask & MVM_EX_CAT_LABELED) {
+            MVMSpeshOperand operand;
+            operand.reg.orig = g->sf->body.handlers[i].label_reg;
+            operand.reg.i = 1;
+            MVM_spesh_usages_add_for_handler_by_reg(tc, g, operand);
+        }
     }
 }
 
