@@ -1151,14 +1151,6 @@ sub emit_names_hash_builder {
     {0x10FFFE,0}
 };
 
-/* Lazily constructed hashtable of Unicode names to codepoints.
-    Okay not to be threadsafe since its value is deterministic
-        and I don't care about the tiny potential for a memory leak
-        in the event of a race condition. */
-
- /* static, so will be 0 initialised. */
-static MVMUniHashTable codepoints_by_name;
-
 static void generate_codepoints_by_name(MVMThreadContext *tc) {
     MVMint32 extent_index = 0;
     MVMint32 codepoint = 0;
@@ -1183,7 +1175,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
                      * <control> <CJK UNIFIED IDEOGRAPH> <CJK COMPATIBILITY IDEOGRAPH>
                      * <surrogate> <TANGUT IDEOGRAPH> <private-use> */
                     if (name && *name != '<') {
-                        MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
+                        MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, name, codepoint);
                     }
                     codepoint++;
                     codepoint_table_index++;
@@ -1192,24 +1184,20 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
             }
             /* Fate NULL */
             case $FATE_NULL:
-                /* codepoint's += ends up being unused since it gets reset at the next loop.
-                 * For now leaving the code as it was. */
-                codepoint += length;
                 break;
             /* Fate Span */
             case $FATE_SPAN: {
                 const char *name = codepoint_names[codepoint_table_index];
                 if (name && *name != '<') {
-                    MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
+                    MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, name, codepoint);
                 }
-                codepoint += length;
                 codepoint_table_index++;
                 break;
             }
         }
     }
     for (; i >= 0; i--) {
-        MVM_uni_hash_insert(tc, &codepoints_by_name, uni_namealias_pairs[i].name, uni_namealias_pairs[i].codepoint);
+        MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, uni_namealias_pairs[i].name, uni_namealias_pairs[i].codepoint);
     }
 
 }
@@ -1308,7 +1296,7 @@ sub emit_unicode_property_keypairs {
         }
     }
     # Make sure General_Category and Script Property values are added first.
-    # These are the only ones (iirc) that are guaranteed in Perl 6.
+    # These are the only ones (iirc) that are guaranteed in Rakudo.
     for my $key (qw(gc sc), sort keys %$PROP_NAMES) {
         for (@{ $aliases{$key} }) {
             next if $PROP_NAMES->{$_};
@@ -1585,7 +1573,7 @@ sub emit_unicode_property_value_keypairs {
             $done{"$propname$_"} ||= push @lines, $lines{$propname}->{$_};
         }
     }
-    my $out = "\nstatic MVMUniHashTable *unicode_property_values_hashes;\n" .
+    my $out = "\n" .
     "static const MVMUnicodeNamedValue unicode_property_value_keypairs[" . scalar(@lines) . "] = {\n" .
     "    " . stack_lines(\@lines, ",", ",\n    ", 0, $WRAP_TO_COLUMNS) . "\n" .
     "};";
@@ -1880,7 +1868,8 @@ sub UnicodeData {
              || $point->{name} eq '<CJK Ideograph Extension D>'
              || $point->{name} eq '<CJK Ideograph Extension E>'
              || $point->{name} eq '<CJK Ideograph Extension F>'
-             || $point->{name} eq '<CJK Ideograph Extension G>')
+             || $point->{name} eq '<CJK Ideograph Extension G>'
+             || $point->{name} eq '<CJK Ideograph Extension H>')
             {
                 $point->{name} = '<CJK UNIFIED IDEOGRAPH>'
             }
@@ -2314,4 +2303,4 @@ sub register_enumerated_property {
 
 main();
 
-# vim: ft=perl6 expandtab sw=4
+# vim: ft=raku expandtab sw=4
